<?php
/**
 * Small example implementation of a Grammar.
 */

require_once __DIR__ . '/Grammar.php';

class ExampleGrammar implements Grammar
{
	private $lexicon;
	private $syntax;

	public function __construct()
	{
		$this->lexicon = $this->getLexicon();
		$this->syntax = $this->getSyntax();
	}

	public function getLexicon()
	{
		return array(
			'a' => array(
				'determiner' => array(),
			),
			'an' => array(
				'determiner' => array(),
			),
			'am' => array(
				'verb' => array(
					'features' => array('head' => array('agreement' => array('person' => 1, 'number' => 's'))))),
			'are' => array(
				'verb' => array(
					'features' => array('head' => array('agreement' => array('person' => 1, 'number' => 'p'))))),
			'author' => array(
				'noun' => array(),
			),
			'born' => array(
				'verb' => array(),
			),
			'by' => array(
				'preposition' => array(),
			),
			'children' => array(
				'noun' => array(),
			),
			'daughter' => array(
				'noun' => array(),
			),
			'did' => array(
				'aux' => array(),
			),
			'die' => array(
				'verb' => array(),
				'roles' => array('patient')
			),
			'flight' => array(
				'noun' => array(
					'features' => array('head' => array('agreement' => array('person' => 3, 'number' => 's')))),
			),
			'have' => array(
				'verb' => array(),
			),
			'how' => array(
				'whword' => array(),
			),
			'i' => array(
				'pronoun' => array(
					'features' => array('head' => array('agreement' => array('person' => 1, 'number' => 's'))))),
			'influenced' => array(
				'verb' => array(),
			),
			'many' => array(
				'determiner' => array(),
			),
			'of' => array(
				'preposition' => array(),
			),
			'the' => array(
				'determiner' => array()
			),
			'that' => array(
				'determiner' => array(),
			),
			'was' => array(
				'aux' => array(),
			),
			'when' => array(
				'whword' => array(),
			),
			'where' => array(
				'whword' => array(),
			),
			'who' => array(
				'whword' => array(),
			),
		);
	}

	/**
	 * Returns all grammar rules with $constituent as their antecedent.
	 *
	 * @param string $constituent
	 * @return array A grammar rule.
	 */
	public function getRulesForConstituent($constituent)
	{
		if (isset($this->syntax[$constituent])) {
			return $this->syntax[$constituent];
		} else {
			return array();
		}
	}

	/**
	 * Returns true if $word belongs to the $partOfSpeech category.
	 *
	 * @param string $word
	 * @param string $partOfSpeech
	 * @return bool
	 */
	public function isWordAPartOfSpeech($word, $partOfSpeech)
	{
		$result = false;

		if (isset($this->lexicon[$word])) {

			if (isset($this->lexicon[$word][$partOfSpeech])) {
				$result = true;
			}

		} else {

			// all words can be proper nouns
			if ($partOfSpeech == 'propernoun') {
				$result = true;
			}
		}

		return $result;
	}

	/**
	 * Returns true if the given syntactic category is a non-abstract part-of-speech.
	 *
	 * @param string $category
	 * @return bool
	 */
	public function isPartOfSpeech($category)
	{
		return in_array($category, array(
			'adjective',
			'adverb',
			'conjunction',
			'determiner',
			'noun',
			'pronoun',
			'numeral',
			'verb',
			'propernoun',
			'whword',
			'aux',
			'preposition',
		));
	}

	/**
	 * Returns the features for a word.
	 * @return array
	 */
	public function getFeaturesForWord($word, $partOfSpeech)
	{
		if (isset($this->lexicon[$word][$partOfSpeech])) {
			if (isset($this->lexicon[$word][$partOfSpeech]['features'])) {
				return $this->lexicon[$word][$partOfSpeech]['features'];
			} else {
				return array();
			}
		} else {
			// presume proper noun
			return array(
				'head' => array(
					'agreement' => array('number' => 's', 'person' => 1),
					'sem' => array('name' => $word)
				)
			);
		}
	}

	public function getSyntax()
	{
		return array(
			'S' => array(
				// John drives
				array(
					array('cat' => 'S', 'features' => array('head-1' => null)),
					array('cat' => 'NP', 'features' => array('head' => array('agreement-2' => null))),
					array('cat' => 'VP', 'features' => array('head-1' => array('agreement-2' => null))),
				),
				array(
					array('cat' => 'S', 'features' => array('head-1' => null)),
					array('cat' => 'WhNP', 'features' => array('head' => array('agreement-2' => null))),
					array('cat' => 'VP', 'features' => array('head-1' => array('agreement-2' => null))),
				),
				array(
					array('cat' => 'S', 'features' => array('head-1' => null)),
					array('cat' => 'VP', 'features' => array('head-1' => null)),
				),
				// Was John driving?
				array(
					array('cat' => 'S', 'features' => array('head-1' => null)),
					array('cat' => 'aux'),
					array('cat' => 'NP', 'features' => array('head' => array('agreement-2' => null))),
					array('cat' => 'VP', 'features' => array('head-1' => array('agreement-2' => null))),
				),
				// Why was John driving?
				array(
					array('cat' => 'S'),
					array('cat' => 'WhNP'),
					array('cat' => 'aux'),
					array('cat' => 'NP'),
					array('cat' => 'VP'),
				),
			),
			'VP' => array(
				array(
					array('cat' => 'VP', 'features' => array('head-1' => null)),
					array('cat' => 'verb', 'features' => array('head-1' => null)),
				),
				array(
					array('cat' => 'VP', 'features' => array('head-1' => null)),
					array('cat' => 'verb', 'features' => array('head-1' => null)),
					array('cat' => 'NP'),
				),
				array(
					array('cat' => 'VP', 'features' => array('head-1' => null)),
					array('cat' => 'verb', 'features' => array('head-1' => null)),
					array('cat' => 'PP'),
				),
			),
			'WhNP' => array(
				array(
					array('cat' => 'WhNP'),
					array('cat' => 'whword'),
				),
				array(
					array('cat' => 'WhNP'),
					array('cat' => 'whword'),
					array('cat' => 'NP'),
				),
			),
			'NP' => array(
				// he
				array(
					array('cat' => 'NP', 'features' => array('head-1' => null)),
					array('cat' => 'pronoun', 'features' => array('head-1' => null)),
				),
				// children
				array(
					array('cat' => 'NP', 'features' => array('head-1' => null)),
					array('cat' => 'noun', 'features' => array('head-1' => null)),
				),
				// John
				array(
					array('cat' => 'NP', 'features' => array('head-1' => null)),
					array('cat' => 'propernoun', 'features' => array('head-1' => null)),
				),
				// the car
				array(
					array('cat' => 'NP', 'features' => array('head-1' => null)),
					array('cat' => 'determiner'),
					array('cat' => 'noun', 'features' => array('head-1' => null)),
				),
				// the car in the lot
				array(
					array('cat' => 'NP', 'features' => array('head-1' => null)),
					array('cat' => 'NP', 'features' => array('head-1' => null)),
					array('cat' => 'PP'),
				),
			),
			'PP' => array(
				// in the lot
				array(
					array('cat' => 'PP', 'features' => array('head-1' => null)),
					array('cat' => 'preposition'),
					array('cat' => 'NP', 'features' => array('head-1' => null)),
				),
			),
		);
	}
}
